// Functions for capitalising a selection.
// Roger Noble, rgn@ukgateway.net

// Version 1.01, 19th May 2000

// History:
// 1.0	19 March 2000	Initial release
// 1.01	19 May 2000	Changed Capitalise menu entry to always return 0 instead of subcode.
//			Previous code caused a type 5 crash sometimes.


void Capitalise_word(int bm)
{
  // Capitalises the word containing the bookmark bm and moves the bookmark to the start
  // of the next word. This routine moves the caret.

  int c;
  int tempbm;
  string s;


  bmmove(bm, 0, 1);       // Move to start of current word

  if (c = bmchar(bm))     // Get first character
  {
    tempbm = bmcreate("Capitalise_tempBM");
    setbmtobm(tempbm, bm);
    bmmove(bm, 1, 5);     // Move bookmark to the start of the next word

    while (bmcmp(tempbm, bm) < 0)
    {
      // Skip punctuation and the like.

      if (c == '(' || c == ')' || c == '[' || c == ']' || c == '{' || c == '}' ||
          c == '\\' || c == '/' || c == '\'' || c == '`' || c == '"' ||
          c == '' || c == '' || c == '' || c == '' ||
          c == '!' || c == '' || c == '?' || c == '' )
      {
        bmmove(tempbm, 1, 0);
        if (!(c = bmchar(tempbm))) break;
        continue;
      }

      s = +chars(c);        // Force into upper case
      setcarettobm(tempbm); // Move caret to start of word

      type("{Deletef}");    // Delete current character
      type(s);              // Replace it with its upper case version
      break;
    }

    bmdelete (tempbm);
  }
}

void Capitalise_selection()
{
  // First check if there is a selection.

  if (zoneorder())
  {

    // There is, go through it and capitalise it.

    int startbm = bmcreate("Capitalise_StartBM");
    int endbm = bmcreate("Capitalise_EndBM");

    setbmtozone(startbm, endbm);
    clearselection();

    while (bmcmp(startbm, endbm) < 0)
      Capitalise_word(startbm);

    setcarettobm(endbm);

    bmdelete(startbm);
    bmdelete(endbm);
  }
  else
  {
    // Nothing is selected, do the word including the current caret position.

    if (caretchar() >= 0)
    {
      int startbm = bmcreate("Capitalise_StartBM");

      setbmtocaret(startbm);

      Capitalise_word(startbm);

      setcarettobm(startbm);	// Move to start of next word

      bmdelete(startbm);
    }
    else
      // No caret, just beep.

      bbc_vdu(7);
  }
}

int Capitalise_menu_entry(int n, int subcode)
{
  if (activetype(TEXTFRAME) <= 1) return 0;

  Capitalise_selection();
  
  return 0;
}

int Capitalise_flags(int entry, string&text)
{
  // Shades or unshades the Capitalise menu entry.

  return ((activetype(TEXTFRAME) > 1) ? 0 : SHADED);
}

void main()
{
  string help = "{CAPITALISE_01}";

  // Do a menu entry for us.

  script_menu_initialise();
  script_menu_insert("Capitalise_menu_entry", "Capitalise_flags", "", "C_\\", "{CAPITALISE_00}");

  // Define a macro as well.

  if (isdefmacro("capitalise") == -1)
  {
    translate(help);
    defmacro3(0, "capitalise", "{if (activetype(TEXTFRAME) > 1) Capitalise_selection(); else bbc_vdu(7)}", "capitalise", help, "Capital");
  }
}

