#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>




#include "wos.h"
#include "fsx.h"
#include "oserr.h"




/* returns 0 ==does not exist  1==file 2==dir */

os_error * fs_exists(char * name,int * type)
{
 os_error   * err;
 DWORD        acc;

 err=XGetFileAttributes(name,&acc);

 if(err)
 {
  *type=0;
  err=NULL;
 }
 else
 {
  *type=(acc & FILE_ATTRIBUTE_DIRECTORY)?2:1;
 }

 return(err);
}




os_error * fs_stat(char * name,fstat * f)
{
 os_error                  * err;
 WIN32_FIND_DATA             fdata;
// arctime                     t;
 HANDLE                      findhandle;
 

 err=XFindFirstFile(name,&fdata,&findhandle);
 if(!err) XFindClose(findhandle);

//err=XGetFileAttributesEx(name,GetFileExInfoStandard,&fdata);
// won't work on '95

 if(!err)
 {
//  osfiletime2riscostime(&t,&fdata.ftLastWriteTime);
//  f->type=fs_filetypefromextension(name);
//  f->load  =f->type<<8;
//  setfiletime(f,&t);
  f->length=fdata.nFileSizeLow;
  f->acc   =fdata.dwFileAttributes;
  f->object=(fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)?2:1;
 }
 else
 {
//  cprintf("fs_stat err");
  err=NULL;
  f->object=0;
 }

 return(err);
}





os_error * fs_read(fshandle fh,void * b,int n)
{
 os_error      * err;
 unsigned long   read;

 err=NULL;

 while(n>0)
 {
  err=XReadFile((HANDLE)fh,b,n,&read,NULL);
  if(!err && read==0)
  {
//   err=geterror(ERD);
   break;
  }
  n-=read;
  b=((char*)b)+read;
 }

 return(err);
}





os_error * fs_loadblock(char *name,void * start)
{
 os_error * err;
 fstat      fs;
 fshandle   fh;

 err=fs_stat(name,&fs);
 if(!err && fs.object==1)
 {
  err=fs_open(name,'r',&fh);
  if(!err)
  {
   err=fs_read(fh,start,fs.length);
   err=fs_close(fh,err);
  }
 }

 return(err);
}




os_error * fs_open(char * name,int mode,fshandle * fh)
{
 os_error * err;
 DWORD      createmode;
 DWORD      accessmode;
 HANDLE     handle;
 int        type;


 if(mode=='r')
 {
  createmode=OPEN_EXISTING;
  accessmode=GENERIC_READ;
 }
 else
 if(mode=='u')
 {
  createmode=OPEN_ALWAYS;
  accessmode=GENERIC_READ|GENERIC_WRITE;
 }
 else
 {
  accessmode=GENERIC_READ|GENERIC_WRITE;
  fs_exists(name,&type);
  if(type==1) createmode=TRUNCATE_EXISTING;
  else
  if(type==0) createmode=CREATE_NEW;
  else
  {
   // error
   createmode=CREATE_NEW;
  }
 }


 err=XCreateFile(name,accessmode,FILE_SHARE_READ,NULL,createmode,FILE_ATTRIBUTE_NORMAL,NULL,&handle);
 if(err) *fh=0;
 else    *fh=(fshandle)handle;

 return(err);
}



os_error * fs_close(fshandle fh,os_error * ep)
{
 os_error * err;

 err=XCloseHandle((HANDLE)fh);

 if(!ep) return(err);
 else    return(ep);
}


os_error * fs_write(fshandle fh,void * b,int n)
{
 os_error      * err;
 unsigned long   written;

 err=NULL;

 while(n)
 {
  err=XWriteFile((HANDLE)fh,b,n,&written,NULL);
  if(err) break;
  n-=written;
  b=(((char*)b)+written);

  if(!err && !written)
  {
//   err=geterror(EWR);
   break;
  }
 }
 return(err);
}




os_error * fs_cdir(char * name)
{
 os_error * err; 
 fstat      f;

 err=NULL;

 fs_stat(name,&f);
 if(f.object==0) err=XCreateDirectory(name,NULL);

 return(err);
}










// for this to work for a directory path has to end with a /
// or the dir flag has to be set to show the last thing in the
// path is a directory

os_error * fs_validatepath(char * path,int dir)
{
 os_error * err;
 char     * p;
 char     * q;
 int        type;
 char       c;


 err=NULL;

// printf("valid %s\n",path);

 p=q=path+strlen(path);

 while(p>path)
 {
  if(*p=='\\' || *p=='/' || (dir && *p==0))
  {
   c=*p;
   *p=0;
   fs_exists(path,&type);
   *p=c;
   if(type)
   {
    p++;
    break;
   }
  }
  p--;
 }

 while(p++<q)
 {
  if(*p=='\\' || *p=='/' || (dir && *p==0))
  {
   c=*p;
   *p=0;
   err=fs_cdir(path);
   if(err) break;
   *p=c;
  }
 }


 return(err);
}





static int xstrnicmp(const char *string1,size_t count1,const char *string2,const char *string3)
{
 size_t count2;
 int    code;

 code=1;

 count2=strlen(string2);
 if(count1==count2)
 {
  code=_strnicmp(string1,string2,count1);
  if(code==0) return(code);
 }

 count2=strlen(string3);
 if(count1==count2)
 {
  code=_strnicmp(string1,string3,count1);
 }

 return(code);
}






static HKEY getrootkey(char * name,int len)
{
 HKEY mkey;

 mkey=NULL;

 if(!xstrnicmp(name,len,"HKEY_CLASSES_ROOT","HKCR")) mkey=HKEY_CLASSES_ROOT;
 else
 if(!xstrnicmp(name,len,"HKEY_CURRENT_USER","HKCU")) mkey=HKEY_CURRENT_USER;
 else
 if(!xstrnicmp(name,len,"HKEY_LOCAL_MACHINE","HKLM")) mkey=HKEY_LOCAL_MACHINE;
 else
 if(!xstrnicmp(name,len,"HKEY_USERS","HKU")) mkey=HKEY_USERS;
 else
 if(!xstrnicmp(name,len,"HKEY_CURRENT_CONFIG","HKCC")) mkey=HKEY_CURRENT_CONFIG;

 return(mkey);
}









static os_error * regclosekey(HKEY * key,os_error * err)
{
 if(*key)
 {
  if(err) XRegCloseKey(*key);
  else    err=XRegCloseKey(*key);
  *key=0;
 }

 return(err);
}





static os_error * regopenkey(char * name,HKEY * key)
{
 os_error * err;
 HKEY       mkey;
 int        len;
 char     * p;

 err=NULL;
 *key=0;

 p=strchr(name,'\\');
 if(p)
 {
  len=p-name;
  mkey=getrootkey(name,len);
  if(mkey)
  {
   p++;

   err=XRegOpenKeyEx((HKEY)mkey,(LPCTSTR)p,0,KEY_ALL_ACCESS,(PHKEY)key);
  }
  else
  {
// err=bad key
  }
 }

 return(err);
}





static os_error * reggetstr(HKEY key,char * name,char * string,int * size)
{
 os_error * err;
 DWORD      ktype;

 err=XRegQueryValueEx(key,name,NULL,&ktype,(BYTE*)string,(DWORD*)size);

 return(err);
}






// here using idea that what comes after last slash is value name
// nb *size is an input/output parameter

os_error * regreadstr(char * keypath,char * string,int * size)
{
 os_error * err;
 HKEY       key;
 char     * p;


 p=strrchr(keypath,'\\');
 if(p) *p=0;
 

 err=regopenkey(keypath,&key);
 if(!err && key)
 {
  if(p)
  {
   err=reggetstr(key,p+1,string,size);
  }
  else
  {
   err=reggetstr(key,NULL,string,size);
  }

  err=regclosekey(&key,err);
 }
 else
 {
  *size=0;
 }

 if(p) *p='\\';

 return(err);
}






