// !Clipboards 1.13
//
// An OvationPro applet that manages the multiple clipboards introduced in OvPro 2.57
//
// Formatted to 100-character width
//
// John Pettigrew
// john-p@innocent.com

int clipboards_total = 1;   // The total number of clipboards in use
int clipboards_status = 0;  // Is the palette open or not?
int clipboards_centre = 1;  // Change this value to 0 to make the palette not be centred by default
int clipboards_palette;     // Holds the window handle
int clipboards_view;
int clipboards_x0;
int clipboards_y1;



// Sets the extent of the palette (i.e. changes its size when the number of clipboards changes)
int clipboards_extent()
{
  int y0;

  y0=geticon(clipboards_palette,clipboards_total*5,1);
  y0-=(geticon(clipboards_palette,5,1)-geticon(clipboards_palette,10,3))/2;


  return(y0);
}

// Run when the palette window is opened and ensures that it is the right size

void clipboards_open3(int handle, int x0, int y0, int x1, int y1, int scx int scy, int bhandle)
{
 if(!clipboards_centre)
 {
  x1+=(clipboards_x0-x0);
  y0+=(clipboards_y1-y1);
  x0=clipboards_x0;
  y1=clipboards_y1;
 }
 else
 {
  clipboards_x0=x0;
  clipboards_y1=y1;
 }

 y0=y1+clipboards_extent();
 open_window(handle,x0,y0,x1,y1,scx,scy,bhandle);
 clipboards_centre=0;
}


void clipboards_open5(int handle, int x0, int y0, int x1, int y1, int scx int scy, int bhandle)
{
 clipboards_x0=x0;
 clipboards_y1=y1;
}


void clipboards_closewindow(void)
{
 close_window(clipboards_palette);
 delete_window(clipboards_palette);
 clipboards_view=0;
 clipboards_palette=0;
}



void clipboards_closepalette(void)
{
 clipboards_closewindow();
}


// Run when the palette closes
void clipboards_close(int handle)
{
 clipboards_closewindow();
 clipboards_status=0;
}


// Set the number of clipboards from the global OvPro choices

void clipboards_setnum(int total)
{
 // Read the number of clipboards in use and set the palette accordingly
 clipboards_total = total;
 if(clipboards_palette)
 {
  writeicon(clipboards_palette, 3, itos(clipboards_total)); //  Set the icon value
 }
}



void clipboards_reset(int total)
{
 clipboards_setnum(total);
 if(clipboards_status) display_window(clipboards_palette, 0, 0); // It must be reopened, of course
}

// Run when the number of clipboards changes
void clipboards_numclipchange(int u, int oldtotal, int newtotal)
{
  // If the total number has changed (i.e. we didn't generate the event by setting the number) then
  // set everything as at startup
 if(clipboards_total!=newtotal) clipboards_reset(newtotal);
}

// Run when the current clipboard changes
void clipboards_currclipchange(int user, int oldclip, int newclip)
{
 int i=54;

 if(clipboards_palette)
 {
  if (newclip > 9) newclip = 0; // If there are more than 10 clipboards, ensure we don't overflow
  while(i++<65) select_icon(clipboards_palette, i, 0); // Unset all radio icons
  select_icon(clipboards_palette, newclip + 55, 1); // Then set the one for the current clipboard
 }
}

// Run when the palette is clicked on
void clipboards_click(int handle, int icon, int bbits, int mx, int my)
{
  int op, current;
  if(icon==101 || icon==102)  // Deal with the bump arrows
  {
    if((icon == 101 && bbits == 4) || (icon == 102 && bbits == 1)) clipboards_total -= 1;
    if((icon == 102 && bbits == 4) || (icon == 101 && bbits == 1)) clipboards_total += 1;
    if(clipboards_total<1) clipboards_total = 1;  // Make sure that the number of clipboards doesn't
    if(clipboards_total>10) clipboards_total = 10;// slip outside the range we can deal with
    generalchoices2(readgeneralchoices2(0), clipboards_total,readgeneralchoices2(2),readgeneralchoices2(3));  // Set the number of boards
    // These next lines change the displayed number of clipbrds
    writeicon(clipboards_palette, 3, itos(clipboards_total));  // and make sure that the size of the
    display_window(clipboards_palette, 0, 0);  // palette changes.
  }
  else  // i.e. the user's clicked on Cut, Copy, Paste or View
  {
    // The operation icons increase in number by 5 for each clipboard (trust me) so taking the mod
    // will let us work out whether the user wants to Cut, Copy, Paste or View
    op=icon % 5;

    // And the clipboards each take up 5 icons
    // so we can work out which clipboard should
    // be used by taking groups of 5 icons.
    // Leaving out the number icons, of course,
    // which are conveniently set up to be
    // numbers 5, 10, 15, 20 etc.

	if(icon<=54)
	{
     current=(icon-5)/5;
	}
	else
    if(icon > 54) // The radio buttons on the left so we only have to change the current clipboard
    {
      current = icon - 55; // Sets the current clipboard
      op = 0; // We're not carrying out any operations, just changing the current clipboard
    }
    setcurrentclipboard(current);  // Now we know which clipboard the user wants, set it
    if(op == 1) cutselection();   // Then use the operation value we worked out above to take the
    if(op == 2) copyselection();  // correct action
    if(op == 3) pasteselection();
    if(op == 4) setclipboard(1);
  }
}



void clipboards_openwindow(int view)
{
 if(view)
 {
  // Create the window and select the first radio icon

  clipboards_palette=create_windowparent("clipboards_pal",viewhandle(view));
  setwindowflags(clipboards_palette,8,8);
  select_icon(clipboards_palette, 55, 1);
  // Read the number of clipboards and set our variable accordingly
  clipboards_setnum(readgeneralchoices2(1));

  //Add handlers for the palette windows
  addwindowhandler(3, clipboards_palette, "clipboards_open3");   // Palette opened
  addwindowhandler(5, clipboards_palette, "clipboards_open5");   // Palette opened
  addwindowhandler(2, clipboards_palette, "clipboards_close");  // Palette closed
  addwindowhandler(0, clipboards_palette, "clipboards_click");  // An icon clicked in the palette

  clipboards_view=view;
 }
}



void clipboards_openpalette(void)
{
 clipboards_openwindow(activeview());
 display_window(clipboards_palette,0,clipboards_centre);
}




void clipboards_activeview(int user,int view,int gain)
{

 if(gain && clipboards_status && view && view!=clipboards_view)
 {
  clipboards_openwindow(view);
  display_window(clipboards_palette,0,clipboards_centre);
 }

 if(!gain && clipboards_status && view && view==clipboards_view && clipboards_palette)
 {
  clipboards_closewindow();
 }

}




// Tick or untick the 'Clipboards' menu entry
int clipboards_tick(int entry, string &text)
{
 return(clipboards_status);
}



// Run when the 'Clipboards' entry in the Applets menu is selected

void clipboards_entry(int entry, int subcode)
{
 clipboards_status=!clipboards_status;
 if(!clipboards_status) clipboards_closepalette();
 else                   clipboards_openpalette();
}



// Main procedure, executed on startup
void main(void)
{
  int i;
  string temp;
  // Set up online help for macro button
  string help = "{CLIPBOARDS_01}";
  translate(help);
  // Initialise
  script_menu_initialise();
  script_menu_insert("clipboards_entry","clipboards_tick","","","{CLIPBOARDS_00}");

  // Add handlers for the total number of clipboards or the current clipboard changing
  addeventhandler(0x136, 0, "clipboards_currclipchange");
  addeventhandler(0x137, 0, "clipboards_numclipchange");
  addeventhandler(0x215, 0, "clipboards_activeview");
  //Create a button macro for the applet that will open the palette
  if(isdefmacro("clipboards") == -1) defmacro3(12,"clipboards","{clipboards_entry(0,0)}","clipboards",help,"clipboards");
}


