/*->c.eps */

/* EPS code */


#include "stdafx.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 
#include <ctype.h>

#include "os.h"
#include "bbc.h"
#include "sprite.h"
#include "wimp.h"
#include "werr.h"

#include "transform.h"

#include "xprocess.h"





#include "scrap.h"
#include "wos.h"
#include "poll.h"
#include "etc.h"
#include "fsx.h"
#include "bf.h"
#include "flex.h"
#include "alloc.h"
#include "xext.h"
#include "task.h"
#include "dbhi.h"
#include "colour.h"
#include "err.h"
#include "filetype.h"
#include "osprint.h"
#include "deb.h"
#include "bits.h"



#include "constants.h"
#include "str.h"

#include "obj.h"
#include "object.h"


#include "file.h"
#include "view.h"

#include "xxcolour.h"
#include "xcolour.h"
#include "colsep.h"
#include "xredraw.h"
#include "geometry.h"
#include "curve.h"

#include "prstr.h"
#include "xprint.h"


#include "pxstr.h"
#include "impex.h"
#include "draw.h"
#include "pixtool.h"
#include "pxtags.h"
#include "pxref.h"
#include "pxcache.h"
#include "transload.h"

#include "dscript.h"








#include "eps.h"



static HANDLE modulehandle;





#define EPSINT   8

#define EPSSHIFT (sizeof(int)*EPSINT)



#define ES100 0x1100
#define ES101 0x1101


static errtag errortable[]=
{
 ES100,"Line in EPS file too long",
 ES101,"DCS included file %s not found",

 ELAST ,""
};

static errlink erroro1;




os_error * epswritedata(int fh,char * data,int len)
{
 return(oswriteprinterdata(data,len));
 USE(fh);
}




/*****************************************************************************/



typedef struct epsobjectstr
{
 int   length;
 int   objectlength;
 int   spare2;
 int   spare3;
 char  name[1];

 /* char data[] */

} epsobjectstr;




typedef struct epshdrstr
{
       int magic;
       int psstart;
       int pslength;
       int metastart;
       int metalength;
       int tiffstart;
       int tifflength;
 short int checksum;

} epshdrstr;


#define EPSMAGIC   0xC6D3D0C5
#define EPSHDRSIZE 30



typedef struct epsparsestr
{
 unsigned int hasboundingbox:1;
 unsigned int hashdr:1;
 unsigned int makeproxy:1;

 int xlo;
 int xhi;
 int ylo;
 int yhi;

 epshdrstr epshdr;

 int    sh;
 int    ftype;

 char * include;
 int    inclen;

} epsparsestr;




static char * epsbuff;
static char * epslinestart;
static char * epslineend;
static char * epsbuffend;

static fshandle    lexfh;

static int    lexeof;

static int    lexmode;


#define LEXBUFFSIZE 0x8000



#define ICHUNK 0x200

static os_error * epsparseinit(epsparsestr * eps)
{
 os_error * err;

 memset(eps,0,sizeof(epsparsestr));

 err=flex_alloc((flex_ptr)&eps->include,ICHUNK);
 if(!err)
 {
  eps->include[0]=0;
 }
 return(err);
}


static os_error * epsparsefinit(epsparsestr * eps)
{
 flex_free((flex_ptr)&eps->include);
 return(NULL);
}


static os_error * epsaddinclude(epsparsestr * eps,char * name)
{
 os_error * err;
 int        len;

 len=strlen(name)+1;

 err=flex_chunk((flex_ptr)&eps->include,eps->inclen+len+1,ICHUNK);
 if(!err)
 {
  strcpy(eps->include+eps->inclen,name);
  eps->inclen+=len;
  eps->include[eps->inclen]=0;
 }

 return(err);
}



static os_error * lexfill(void)
{
 os_error * err;
 int n;
 int m;


 m=epslinestart-epsbuff;
 n=epsbuffend-epslinestart;

 if(n>=LEXBUFFSIZE) err=geterror(ES100); /* error line too long */
 else
 {
  epslineend-=m;

  memmove(epsbuff,epslinestart,n);
  epslinestart=epsbuff;
  epsbuffend=epsbuff+n;
  m=LEXBUFFSIZE-n;

  err=fs_readn(lexfh,epsbuff+n,m,&n);
  epsbuffend+=n;
  lexeof=n<m;
 }

 return(err);
}



static os_error * epsstart(char * name,char * data,int len,epsparsestr * eps)
{
 os_error * err;
 int        n;
 fshandle   fh;
 int        length;
 int        read;
 int        pstart;
 int        plength;

 err=NULL;

 pstart=plength=0; /* compiler */


 if(name)
 {
  err=fs_open(name,'r',&lexfh);
  if(!err)
  {
   err=salloc((flex_ptr)&epsbuff,LEXBUFFSIZE+1+3);
   if(!err)
   {
    err=fs_readn(lexfh,epsbuff,4,&n);
    if(!err && n==4)
    {
     if(*((int*)epsbuff)==EPSMAGIC)
     {
      XMEMCPYX(&eps->epshdr.magic,epsbuff,4);

      eps->hashdr=1;


      err=fs_read(lexfh,&eps->epshdr.psstart,EPSHDRSIZE-4);
      if(!err)
      {
      /* dump out tiff image to scrap file */


       if(eps->makeproxy)
       {
        if(eps->epshdr.tifflength)
        {
         pstart=eps->epshdr.tiffstart;
         plength=eps->epshdr.tifflength;
         eps->ftype=TIFFTYPE;
        }
        else
        if(eps->epshdr.metalength)
        {
         pstart=eps->epshdr.metastart;
         plength=eps->epshdr.metalength;
         eps->ftype=WMETAF;
        }


        if(plength)
        {
         err=sc_create(&eps->sh);
         if(!err) err=fs_seek(lexfh,pstart);
         if(!err) err=fs_open(sc_name(eps->sh),'w',&fh);
         if(!err)
         {

          length=plength;
          while(length)
          {
           read=MIN(LEXBUFFSIZE,length);
           err=fs_read(lexfh,epsbuff,read);
           if(err) break;
           err=fs_write(fh,epsbuff,read);
           if(err) break;
           length-=read;
          }
          err=fs_close(fh,err);
          if(!err) err=fs_settype(sc_name(eps->sh),eps->ftype);
         }
        }
       }


       if(!err) err=fs_seek(lexfh,eps->epshdr.psstart);
      }
      n=0;
     }
    }

    if(!err)
    {
     epslinestart=epslineend=epsbuff;
     epsbuffend=epsbuff+n;
     lexeof=0;
     lexmode=0;
   

     err=lexfill();
    }

    if(err) sfree((flex_ptr)&epsbuff);
   }

   if(err) err=fs_close(lexfh,err);
  }
 }
 else
 {
  if(len>4 && *((int*)data)==EPSMAGIC)
  {
   eps->hashdr=1;

   XMEMCPYX(&eps->epshdr,data,EPSHDRSIZE);

   if(eps->makeproxy)
   {
    if(eps->epshdr.tifflength)
    {
     pstart=eps->epshdr.tiffstart;
     plength=eps->epshdr.tifflength;
     eps->ftype=TIFFTYPE;
    }
    else
    if(eps->epshdr.metalength)
    {
     pstart=eps->epshdr.metastart;
     plength=eps->epshdr.metalength;
     eps->ftype=WMETAF;
    }

    if(plength)
    {
     err=sc_create(&eps->sh);
     if(!err) err=fs_saveblock(sc_name(eps->sh),eps->ftype,
                                                data+pstart,plength);
    }
   }

   epslinestart=epslineend=epsbuff=data+eps->epshdr.psstart;
   /* epsbuffend=epsbuff+eps->epshdr.pslength; */
   epsbuffend=MIN((data+len),(epsbuff+eps->epshdr.pslength));
  }
  else
  {
   epslinestart=epslineend=epsbuff=data;
   epsbuffend=epsbuff+len;
  }

  lexeof=1;
  lexmode=1;
 }


 return(err);
}




static os_error * epsend(os_error * err)
{

 err=NULL;

 if(lexmode==0)
 {
  err=fs_close(lexfh,err);
  sfree((flex_ptr)&epsbuff);
 }

 return(err);
}




static os_error * epsskip(int len)
{
 os_error * err;
 int        n;
 int        posn;

 err=NULL;

 epslinestart=epslineend;
 epslinestart+=len;
 epslineend=epslinestart;

 if(epslinestart>=epsbuffend)
 {
  if(!lexeof)
  {
            err=fs_tell(lexfh,&posn);
   if(!err) err=fs_seek(lexfh,posn+epslinestart-epsbuffend);
   if(!err)
   {
    err=fs_readn(lexfh,epsbuff,LEXBUFFSIZE,&n);
    epsbuffend=epsbuff+n;
    lexeof=n<LEXBUFFSIZE;
    epslinestart=epslineend=epsbuff;
   }
  }
 }

 return(err);
}





static os_error * epscopy(int len,int fh)
{
 os_error * err;
 int        read;

 err=NULL;

 epslinestart=epslineend;

 while(len)
 {
  read=MIN(len,(epsbuffend-epslinestart));
  cprintf("epscopy %d",read);
  err=epswritedata(fh,epslinestart,read);
  if(err) break;
  len-=read;
  epslinestart+=read;
  epslineend=epslinestart;

  if(epslineend>=epsbuffend)
  {
   if(lexeof) break;
   else       err=lexfill();
  }
  if(err) break;
 }

 return(err);
}






static os_error * epssetline(void)
{
 os_error * err;
 int        c;

 err=NULL;

 while(1)
 {
  c=*epslineend;
  if(c=='\n' || c=='\r') break;
  epslineend++;
  if(epslineend>=epsbuffend)
  {
   if(lexeof) break;
   else       err=lexfill();
   if(err)    break;
  }
 }


 if(!err && epslineend<epsbuffend)
 {
  while(1)
  {
   c=*epslineend;
   if(c!='\n' && c!='\r') break;
   epslineend++;
   if(epslineend>=epsbuffend)
   {
    if(lexeof) break;
    else       err=lexfill();
    if(err)    break;
   }
  }
 }

 return(err);
}






static os_error * epsgetline(void)
{
 os_error * err;

 err=NULL;

 epslinestart=epslineend;

 if(!err && epslinestart<epsbuffend) err=epssetline();

 return(err);
}



#define TKEOL          0x0
#define TKPCPC         0x1
#define TKBBOX         0x2
#define TKBEGINBIN     0x3
#define TKINT          0x4
#define TKBEGINDOC     0x5
#define TKENDDOC       0x6
#define TKATEND        0x7
#define TKCYANP        0x8
#define TKMAGENTAP     0x9
#define TKYELLOWP      0xA
#define TKBLACKP       0xB
#define TKKEYP         0xC
#define TKBEGINCUSTOM  0xD
#define TKENDCUSTOM    0xE
#define TKBEGINPROCESS 0xF
#define TKBEGINPREVIEW 0x10
#define TKENDPREVIEW   0x11
#define TKBEGINDATA    0x12
#define TKENDDATA      0x13
#define TKHEX          0x14
#define TKBINARY       0x15
#define TKASCII        0x16
#define TKLINES        0x17
#define TKBYTES        0x18
#define TKSTR          0x19



static int    current;
static int    currentval;
static char * tokptr;


typedef struct tokstr
{
 char * token;
 int    val;
} tokstr;


static tokstr tokens[]=
{
 "%%",               TKPCPC,
 "BoundingBox:",     TKBBOX,
 "BeginBinary:",     TKBEGINBIN,
 "BeginDocument:",   TKBEGINDOC,
 "EndDocument",      TKENDDOC,
 "(atend)",          TKATEND,
 "CyanPlate:",       TKCYANP,
 "MagentaPlate:",    TKMAGENTAP,
 "YellowPlate:",     TKYELLOWP,
 "BlackPlate:",      TKBLACKP,
 "KeyPlate:",        TKKEYP,
 "BeginCustomColor", TKBEGINCUSTOM,
 "EndCustomColor",   TKENDCUSTOM,
 "BeginProcessColor",TKBEGINPROCESS,
 "EndProcessColor",  TKBEGINPROCESS,
 "BeginPreview",     TKBEGINPREVIEW,
 "EndPreview",       TKENDPREVIEW,
 "BeginData",        TKBEGINDATA,
 "EndData",          TKENDDATA,
 "Hex",              TKHEX,
 "Binary",           TKBINARY,
 "ASCII",            TKASCII,
 "Lines",            TKLINES,
 "Bytes",            TKBYTES,

 NULL,0,
};



static os_error * epssettoken(void)
{
 current=TKEOL;

 if(epslineend>(epslinestart+2))
 {
  tokptr=epslinestart;
  if(*tokptr++=='%' && *tokptr++=='%') current=TKPCPC;
 }

 return(NULL);
}



static int tkstrcmp(char * p,char * q,char * qmax)
{

 while(q<qmax)
 {
  if(*p==0)  return(0);
  if(*p!=*q) return(1);

  p++;
  q++;
 }

 return(1);
}



static os_error * epsgettoken(void)
{
 int    i;
 char * p;
 int    c;
 int    minus;



 current=TKEOL;

 while(tokptr<epslineend)
 {
  if(*tokptr>' ') break;
  tokptr++;
 }

 if(tokptr<epslineend)
 {
  current=TKSTR;

  c=*tokptr;

  if((c>='0' && c<='9') || c=='-')
  {
   if(c=='-') {currentval=0;minus=1;}
   else       {currentval=c-'0';minus=0;}
   while(1)
   {
    tokptr++;
    if(tokptr>=epslineend) break;
    c=*tokptr;

    if(c>='0' && c<='9') currentval=currentval*10+(c-'0');
    else                 break;
   }

   if(minus) currentval=-currentval;

   current=TKINT;
  }
  else
  {
   i=0;

   while(1)
   {
    p=tokens[i].token;
    if(!p) break;

    if(!tkstrcmp(p,tokptr,epslineend))
    {
     current=tokens[i].val;
     tokptr+=strlen(tokens[i].token);
     break;
    }
    i++;
   }
  }
 }


 return(NULL);
}



static os_error * epsgetstring(int stringmax,char * string)
{
 int    c;
 char * lt;

 lt=string+stringmax-1;

 current=TKEOL;

 while(tokptr<epslineend)
 {
  if(*tokptr>' ') break;
  tokptr++;
 }

 if(tokptr<epslineend)
 {
  current=TKSTR;

  while(tokptr<epslineend)
  {
   c=*tokptr++;
   if(c<=' ') break;
   *string++=c;
   if(string>=lt) break;
  }
 }

 *string=0;

 return(NULL);
}








static os_error * epsparse(epsparsestr * eps,char * name,char * data,int len)
{
 os_error * err;
 int        indoc;
 int        getline;
 int        n;
 char       temp[256];

 indoc=0;



 err=epsstart(name,data,len,eps);
 if(!err)
 {
  epssetline();

  while(epslinestart<epsbuffend)
  {
   getline=1;
   epssettoken();

   if(current==TKPCPC)
   {
    epsgettoken();

    switch(current)
    {
     case     TKBBOX:
                     if(indoc) break;

                     epsgettoken();
                     if(current!=TKINT) break;
                     eps->xlo=currentval;

                     epsgettoken();
                     if(current!=TKINT) break;
                     eps->ylo=currentval;

                     epsgettoken();
                     if(current!=TKINT) break;
                     eps->xhi=currentval;

                     epsgettoken();
                     if(current!=TKINT) break;
                     eps->yhi=currentval;

                     eps->hasboundingbox=1;

                     break;


     case TKBEGINDOC:
                     indoc++;
                     break;

     case   TKENDDOC:
                     indoc--;
                     break;

     case TKBEGINBIN:
                     epsgettoken();
                     if(current==TKINT)
                     {
                      epsskip(currentval);
                      epssetline();
                      getline=0;
                     }
                     break;

     case TKBEGINDATA:
                      epsgettoken();

                      if(current!=TKINT) break;

                      n=currentval;

                      epsgettoken();

                      if(current!=TKHEX && 
                         current!=TKBINARY && current!=TKASCII) break;

                      epsgettoken();

                      if(current!=TKBYTES && current!=TKLINES) break;

                      if(current==TKBYTES) 
                      {
                       epsskip(n);
                       epssetline();
                       getline=0;
                      }
                      else
                      if(current==TKLINES)
                      {
                       while(n--)
                       {
                        epsgetline();
                        if(epslinestart>=epsbuffend) break;
                       }
                      }
                      break;

    case      TKCYANP:
    case   TKMAGENTAP:
    case    TKYELLOWP:
    case     TKBLACKP:
    case       TKKEYP:
                      epsgetstring(NITEM(temp),temp);
                      if(current==TKSTR)
                      {
                       epsaddinclude(eps,temp);
                      }
                      break;

    }
   }

   if(err) break;

   if(getline) epsgetline();
  }

  err=epsend(err);
 }

 return(err);
}





static os_error * epsmkproxy(epsparsestr * eps,int picture,filestr * file,char * name)
{
 os_error   * err;
 char         from[FILENAMELEN];
 char         to[FILENAMELEN];
 int          result;  
 int          sh;
 fstat        fs;
 int          code;
 int          needproxy;
 int          x0;
 int          x1;
 int          y0;
 int          y1;


 err=NULL;

 if(eps->hasboundingbox)
 {
  x0=eps->xlo*640;
  y0=eps->ylo*640;
  x1=eps->xhi*640;
  y1=eps->yhi*640;
 }
 else
 {
  x0=0;  /* low x */
  y0=0;  /* low y */
  x1=(49*256*1000)/400;  /* hi  x */
  y1=(43*256*1000)/400;  /* hi  y */
 }
 
 needproxy=1;
 
 if(name)
 {
  sc_create(&sh);
  xzstrncpy(to,sc_name(sh),NITEM(to));

  if(eps->hashdr && eps->sh) xzstrncpy(from,sc_name(eps->sh),sizeof(from));
  else                       *from=0;

  err=xexec("eps_makeproxy",&result,"SSIIIIS/R",name,from,x0,y0,x1,y1,to,&code);

  if(!err && code==0 && result==0)
  {
   err=fs_stat(to,&fs);
   if(!err && fs.object)
   {
    err=sploadproxy(to,picture,file);
    needproxy=0;
   }
  }

  fs_delete(to);
  sc_remove(sh,NULL);
 }


 if(eps->hashdr)
 {
  if(eps->sh)
  {
   if(needproxy)
   {
    xzstrncpy(from,sc_name(eps->sh),sizeof(from));
    sc_create(&sh);
    xzstrncpy(to,sc_name(sh),sizeof(to));

    err=xexec("eps_convert",&result,"SIS/R",from,eps->ftype,to,&code);

    if(!err && code==0 && result==0)
    {
     err=fs_stat(to,&fs);
     if(!err && fs.object)
     {
      err=drloadproxy(to,picture,file);
     }
    }

    fs_delete(to);
    fs_delete(from);
//			cprintf("from %s",from);
    sc_remove(sh,NULL);
   }
   sc_remove(eps->sh,NULL);
  }
 }

 return(err);
}




static void fixname(char * name,char * table)
{
 int    c;
 char * p;

 while((c=*name)!=0)
 {
  p=table;
  while(*p)
  {
   if(*p==c)
   {
    *name=*(p+1);
    break;
   }
   p+=2;
  }
  name++;
 }
}



static os_error * epscheckfile(char * path,char * leaf,char * name)
{
 os_error * err;
 fstat      fs;


 strcpy(leaf,name);

 err=fs_stat(path,&fs);
 if(!err)
 {

  if(!fs.object)
  {
   err=geterrorf(ES101,name);  /* inc file missing */
  }


#ifdef NEVER
  if(!fs.object)
  {
   fixname(leaf,"./");

   err=fs_stat(path,&fs);
   if(!err)
   {
    if(!fs.object)
    {
     strcpy(leaf,name);
     fixname(leaf,".");
     err=fs_stat(path,&fs);
     if(!err)
     {
      if(!fs.object)
      {
       err=geterrorf(ES101,name);  /* inc file missing */
      }
     }
    }
   }
  }
#endif
 }


 return(err);
}




static os_error * epscheckincludedfiles(epsparsestr * eps,char * name)
{
 os_error     * err;
 int            current;
 char           path[FILENAMELEN];
 char         * leaf;
 int            namelen;


 err=NULL;

 current=0;

 xzstrncpy(path,name,sizeof(path));
 leaf=fs_leaf(path);

 while((namelen=strlen(eps->include+current))!=0)
 {
  err=epscheckfile(path,leaf,eps->include+current);
  if(err) break;

  current+=namelen+1;
 }

 return(err);
}








static os_error * epsembedincludedfiles(epsparsestr * eps,char * name,
                                              int picture,filestr * file)
{
 os_error     * err;
 int            current;
 char           path[FILENAMELEN];
 char         * leaf;
 int            len;
 int            namelen;
 fstat          fs;
 picturestr   * picturep;
 int            imlen;
 int            hdrlen;
 int            insert;
 epsobjectstr * epsobj;



 err=NULL;



 current=0;

 xzstrncpy(path,name,sizeof(path));
 leaf=fs_leaf(path);


 while((namelen=strlen(eps->include+current))!=0)
 {
  err=epscheckfile(path,leaf,eps->include+current);
  if(err) break;

  err=fs_stat(path,&fs);
  if(err) break;

  picturep=(picturestr*)fob_find(picture,file);


  imlen=(picturep->imagelen+3) & (~0x3);

  len=imlen+picturep->proxylen+picturep->tagslen+sizeof(picturestr);

  hdrlen=sizeof(epsobjectstr)+strlen(eps->include+current)+1;
  hdrlen=(hdrlen+3) & (~0x3);

  insert=(hdrlen+fs.length+3)&(~0x3);

  err=fob_change(picture,file,insert+len,(fobjptr)&picturep);
  if(!err)
  {
   memmove(picturep->data+imlen+insert,
           picturep->data+imlen,
           picturep->proxylen+picturep->tagslen);

   epsobj=(epsobjectstr*)(picturep->data+imlen);
   memset(epsobj,0,sizeof(epsobjectstr));
   epsobj->objectlength=fs.length;
   epsobj->length=insert;
   strcpy(epsobj->name,eps->include+current);


   err=fs_loadblock(path,picturep->data+imlen+hdrlen);
   if(!err)
   {
    picturep->imagelen=imlen+insert;
   }

  }

  if(err) break;

  current+=namelen+1;
 }


 return(err);
}








static void hdrshift(picturestr  * picturep,int * shift,int * length)
{
 epsobjectstr * epsobj;
 int            hdrlen;

 epsobj=(epsobjectstr*)(picturep->data+EPSSHIFT);
 hdrlen=sizeof(epsobjectstr)+strlen(epsobj->name)+1;
 hdrlen=(hdrlen+3) & (~0x3);

 if(shift)  *shift=(EPSSHIFT+hdrlen);
 if(length) *length=epsobj->objectlength;
}




static os_error * epsfindincluded(char * name,picturestr * picturep,
                                  char ** data,int * len)
{
 os_error     * err;
 epsobjectstr * epsobj;
 int            offset;
 int            hdrlen;

 err=NULL;

 offset=EPSSHIFT;

 while(offset<picturep->imagelen)
 {
  epsobj=(epsobjectstr*)(picturep->data+offset);

  if(strcmp(name,epsobj->name)==0)
  {
   hdrlen=sizeof(epsobjectstr)+strlen(epsobj->name)+1;
   hdrlen=(hdrlen+3) & (~0x3);
   *data=(picturep->data+offset+hdrlen);
   *len=epsobj->objectlength;
   break;
  }

  offset+=epsobj->length;
 }

 if(offset>=picturep->imagelen) err=geterrorf(ES101,name);

 return(err);
}






static os_error * initpicture(int picture,filestr * file,
                              int inmemory,char * name,epsparsestr * eps)
{
 os_error     * err;
 picturestr   * picturep;
 int            shift;
 int            length;


 eps->makeproxy=1;

 picturep=(picturestr*)fob_find(picture,file);

 if(!inmemory)
 {
  err=epsparse(eps,name,NULL,0);  /* DP 329 */
 }
 else
 {
  hdrshift(picturep,&shift,&length);
  err=epsparse(eps,name,picturep->data+shift,length);
 }

 if(!err)
 {
  picturep=(picturestr*)fob_find(picture,file);

  if(eps->hasboundingbox)
  {
   picturep->abox.x0=eps->xlo*640;
   picturep->abox.y0=eps->ylo*640;
   picturep->abox.x1=eps->xhi*640;
   picturep->abox.y1=eps->yhi*640;
  }
  else
  {
   picturep->abox.x0=0;  /* low x */
   picturep->abox.y0=0;  /* low y */
   picturep->abox.x1=(49*256*1000)/400;  /* hi  x */
   picturep->abox.y1=(43*256*1000)/400;  /* hi  y */
  }
 }

 if(!err) err=epsmkproxy(eps,picture,file,name);

 return(err);
}




static os_error * epsimportpost(int ftype,int object,filestr * file)
{
 USE(ftype);
 USE(object);
 USE(file);

 return(NULL);
}



static os_error * epsimport(char * name,int ftype,int xv,
                                                int * object,filestr * file)
{
 os_error     * err;
 picturestr   * picturep;
 int            picture;
 int            hdrlen;
 epsobjectstr * epsobj;
 epsparsestr    eps;

 hdrlen=sizeof(epsobjectstr)+strlen(name)+1;
 hdrlen=(hdrlen+3) & (~0x3);

 err=dxloadchunk(name,EPSSHIFT+hdrlen,object,file);
 if(!err)
 {
  picture=*object;
  picturep=(picturestr*)fob_find(picture,file);
  picturep->ftype=ftype;

  memset(picturep->data,0,EPSSHIFT);
  epsobj=(epsobjectstr*)(picturep->data+EPSSHIFT);
  memset(epsobj,0,sizeof(epsobjectstr));
  epsobj->objectlength=picturep->imagelen-(EPSSHIFT+hdrlen);
  epsobj->length=(hdrlen+epsobj->objectlength+3) & (~0x3);
  strcpy(epsobj->name,name);


  err=epsparseinit(&eps);
  if(!err)
  {
   err=initpicture(picture,file,1,xv?NULL:name,&eps);
   if(!err) err=epsembedincludedfiles(&eps,name,picture,file);
   epsparsefinit(&eps);
  }

  if(err) dxscrapchunk(object,file);
 }

 return(err);

 USE(xv);
}



static os_error * epswriteline(int fh)
{
 os_error * err;

// cprintf("wln %d",epslineend-epslinestart);

 err=epswritedata(fh,epslinestart,epslineend-epslinestart);

 return(err);
}







static os_error * epsparseplate(char * name,char * data,int len,
                                int platemax,char * plate,int sep)
{
 os_error     * err;
 epsparsestr    eps;
 int            indoc;
 int            getline;
 int            n;


 *plate=0;

 indoc=0;

 epsparseinit(&eps);

 err=epsstart(name,data,len,&eps);
 if(!err)
 {
  epssetline();

  while(epslinestart<epsbuffend)
  {
   getline=1;
   epssettoken();

   if(current==TKPCPC)
   {
    epsgettoken();

    switch(current)
    {

      case     TKBEGINDOC:
                          indoc++;
                          break;

      case       TKENDDOC:
                          indoc--;
                          break;


      case     TKBEGINBIN:
                          epsgettoken();
                          if(current==TKINT)
                          {
                           epsskip(currentval);
                           epssetline();
                           getline=0;
                          }
                          break;

     case     TKBEGINDATA:
                          epsgettoken();

                          if(current!=TKINT) break;

                          n=currentval;

                          epsgettoken();

                          if(current!=TKHEX && 
                             current!=TKBINARY && current!=TKASCII) break;

                          epsgettoken();

                          if(current!=TKBYTES && current!=TKLINES) break;

                          if(current==TKBYTES) 
                          {
                           epsskip(n);
                           epssetline();
                           getline=0;
                          }
                          else
                          if(current==TKLINES)
                          {
                           while(n--)
                           {
                            epsgetline();
                            if(epslinestart>=epsbuffend) break;
                           }
                          }
                          break;


        case      TKCYANP:
                          if(sep==KEYCYAN) epsgetstring(platemax,plate);
                          break;

        case   TKMAGENTAP:
                          if(sep==KEYMAGENTA) epsgetstring(platemax,plate);
                          break;

        case    TKYELLOWP:
                          if(sep==KEYYELLOW) epsgetstring(platemax,plate);
                          break;

        case     TKBLACKP:
        case       TKKEYP:
                          if(sep==KEYBLACK)  epsgetstring(platemax,plate);
                          break;

                  default:
                          break;
    }
   }

   if(err) break;

   if(getline) epsgetline();
  }

  err=epsend(err);
 }

 epsparsefinit(&eps);

 return(err);
}








static os_error * epsrenderdata2(int fh,char * name,char * data,int len)
{
 os_error     * err;
 epsparsestr    eps;
 int            indoc;
 int            skipdata;
 int            getline;
 int            n;


 indoc=skipdata=0;

 epsparseinit(&eps);

 err=epsstart(name,data,len,&eps);
 if(!err)
 {
  epssetline();

  while(epslinestart<epsbuffend)
  {
   getline=1;
   epssettoken();

   if(current==TKPCPC)
   {
    epsgettoken();

    switch(current)
    {

      case TKBEGINPREVIEW:
                          skipdata=1;
                          break;


      case   TKENDPREVIEW:
                          skipdata=0;
                          break;


      case     TKBEGINDOC:
                          indoc++;
                          err=epswriteline(fh);
                          break;


      case       TKENDDOC:
                          indoc--;
                          err=epswriteline(fh);
                          break;


      case     TKBEGINBIN:
                          epsgettoken();
                          if(current==TKINT)
                          {
                                    err=epswriteline(fh);
                           if(!err) err=epscopy(currentval,fh);
                           epssetline();
                           getline=0;
                          }
                          break;


      case    TKBEGINDATA:
                          epsgettoken();

                          if(current!=TKINT) break;

                          n=currentval;

                          epsgettoken();
                         
                          if(current!=TKHEX && 
                            current!=TKBINARY && current!=TKASCII) break;

                          epsgettoken();

                          if(current!=TKBYTES && current!=TKLINES) break;

                          err=epswriteline(fh);
                          if(err) break;

                          if(current==TKBYTES) 
                          {
                           err=epscopy(n,fh);
                           epssetline();
                           getline=0;
                          }
                          else
                          if(current==TKLINES)
                          {
                           while(n-->0)
                           {
                            epsgetline();
                            if(epslinestart>=epsbuffend) break;
                            err=epswriteline(fh);
                            if(err) break;
                           }
                          }
                          break;


                  default:
                          if(!skipdata) err=epswriteline(fh);
                          break;
    }
   }
   else
   {
    err=epswriteline(fh);
   }

   if(err) break;

   if(getline) epsgetline();
  }

  err=epsend(err);
 }

 epsparsefinit(&eps);


 return(err);
}




static os_error * epsrenderdata(transformstr * tr,processstr * pr,
                                int fh,picturestr * picturep,filestr * file)
{
 os_error     * err;
 char           expandedname[FSMAXPATH];
 char         * name;
 char         * data;
 int            picture;
 int            len;
 int            shift;
 int            result;
 epsobjectstr * epsobj;
 char           temp[256];
 int            sep;
 char         * leaf;
 int            ftype;
 int            ptype;
 int            transhandle;
 char           transname[FILENAMELEN];
 void         * eh;



 err=NULL;

 picture=picturep->key;


 if(picturep->imagelen==0)
 {
  ftype=picturep->ftype;
  err=pxrendername(picturep,file,expandedname,NITEM(expandedname));
  name=expandedname;
  data=NULL;
  len=0;

  ptype=fs_filetypefromextension(expandedname);
  if(ptype!=ftype)
  {
   err=sc_create(&transhandle);
   if(!err)
   {
    strcpy(transname,sc_name(transhandle));
    name=transname;
    err=translatefile(expandedname,-1,transname,ftype);
   }
  }
 }
 else
 {
  name=NULL;
  hdrshift(picturep,&shift,&len);
  data=picturep->data+shift;
  epsobj=(epsobjectstr*)(picturep->data+EPSSHIFT);
  err=xzstrncpyu(expandedname,epsobj->name,NITEM(expandedname));
 }


	if(!err)
	{
  sep=pr->separation;

  if(sep==KEYCYAN || sep==KEYMAGENTA || sep==KEYBLACK || sep==KEYYELLOW)
  {
   err=epsparseplate(name,data,len,NITEM(temp),temp,sep);
   if(!err)
   {
    if(*temp)
    {
     if(picturep->imagelen==0)
     {
      leaf=fs_leaf(expandedname);
      err=epscheckfile(expandedname,leaf,temp);
      if(!err)
      {
       name=expandedname;
      }
     }
     else
     {
      err=epsfindincluded(temp,picturep,&data,&len);
      if(!err)
      {
       err=xzstrncpyu(expandedname,temp,NITEM(expandedname));
      }
     }

     sep=0;
    }
   }
  }
	}

 if(!err)
 {

/*  xexec("eps_prerender",&result,"IIIIIISI",
                  tr->a,tr->b,tr->c,tr->d,tr->e,tr->f,expandedname,sep); */

  xexec("eps_prerender",&result,"IIIIIISI",
                  tr->a,tr->b,tr->c,tr->d,tr->e,tr->f,expandedname,sep);

  err=epsrenderdata2(fh,name,data,len);


  xexec("eps_postrender",&result,"I",sep);
 }

 if(name==transname)
 {
  eh=errsave(err);
  fs_delete(sc_name(transhandle));
  sc_remove(transhandle,err);
  err=errload(eh,NULL);
 }

 return(err);
}




/* render picture                */
/* given box in picture space    */
/* given transform to real space */

static os_error * epsrender(aboxstr * abox,transformstr * tr,
                     processstr * pr,picturestr * picturep,viewredrawstr * v)
{
 os_error * err;
 prjobstr * prjob;
 int        path[20];
 pathstyle  style;
 curvestr   c;

 err=NULL;

 if(v->isprinter)
 {
  prjob=(prjobstr*)v->prjob;
  
  cprintf("epsrender type *** %d",prjob->prtype);

  if(prjob->prtype==PRPS && (prjob->bits & PSDIRECT))
  {
   err=epsrenderdata(tr,pr,prjob->newprinthandle,picturep,v->file);
  }
  else
  {
   if(!pr->useproxies && picturep->proxylen)
   {
    pr->useproxies=1;
    err=render(abox,tr,pr,picturep,v);
    pr->useproxies=0;
   }
  }
 }
 else
 {
  abox2box(&picturep->abox,&c.box);
  c.type=CBOX;

  rcreatepathstyle(&style,0);

  curvepath(&c,path);

  style.action=0x2;
  rwimp_setcolour(WIMPGREY1);
  rpathlines(path,&style,v,tr);

  style.action=0x1;
  rwimp_setcolour(WIMPBLACK);
  rpathlines(path,&style,v,tr);
 }


 return(err);

 USE(abox);
 USE(pr);
}



static os_error * epsdsexport(int ftype,int object,
                                           filestr * file,buffer * bf)
{
 picturestr * picturep;
 int          len;

 picturep=(picturestr*)fob_find(object,file);

 len=picturep->imagelen-EPSSHIFT;

 bf_printfeol(bf,"%d",len);

 dxwritedata(NULL,-1,bf);

 dxwritedata(picturep->data+EPSSHIFT,len,bf);

 return(dxwritedata(NULL,-2,bf));

 USE(ftype);
}





static os_error * epsdsimport(int ftype,int fp,int argc,objectloc * loc)
{              
 os_error          * err;
 filestr           * file;
 dsobjectstr       * ds;
 int                 object;
 picturestr        * picturep;
 epsparsestr         eps;

 err=NULL;

 file=loc->file;

 dxloadblock(EPSSHIFT,fp,argc,loc);

 if(argc==DSIMMLAST)
 {
  ds=(dsobjectstr*)(dsstackptr()+fp);
  object=ds->data.integer;

  picturep=(picturestr*)fob_find(object,loc->file);
  picturep->ftype=ftype;

  err=epsparseinit(&eps);
  if(!err)
  {
   err=initpicture(object,file,1,NULL,&eps);
   epsparsefinit(&eps);
  }

  if(err) 
  {
   dxscrapchunk(&object,file);
   ds=(dsobjectstr*)(dsstackptr()+fp);
   ds->data.integer=object;
  }
 }

 return(err);
}



static os_error * epsexport(char * name,int object,int ftype,int xv,
                                                              filestr * file)
{
 os_error   * err;
 picturestr * picturep;
 int          shift;
 int          length;

 picturep=(picturestr*)fob_find(object,file);
 hdrshift(picturep,&shift,&length);
 err=bf_saveblock(name,ftype,picturep->data+shift,length);

 return(err);

 USE(xv);
}



static int epsdoseps;

static os_error * epsprcache0(int eventn,int userhandle,
                                         viewstr * view,prjobstr * prjob)
{

 epsdoseps=prjob->doseparations;

 return(NULL);

 USE(userhandle);
 USE(eventn);
 USE(view);
 USE(prjob);
}





static os_error * epscache(picturestr * picturep,filestr * file)
{
 os_error     * err;
 char           expandedname[FSMAXPATH];
 int            len;
 int            shift;
 epsobjectstr * epsobj;
 int            result;


 err=NULL;

 if(picturep->imagelen==0)
 {
  err=pxrendername(picturep,file,expandedname,NITEM(expandedname));
 }
 else
 {
  hdrshift(picturep,&shift,&len);
  epsobj=(epsobjectstr*)(picturep->data+EPSSHIFT);
  err=xzstrncpyu(expandedname,epsobj->name,NITEM(expandedname));
 }

 if(!err)
 {
  cprintf("eps cache");

  xexec("eps_prcache",&result,"SI",expandedname,epsdoseps);
 }

 return(err);
}



static int epsrendermem(picturestr * picturep,transformstr * tr,
                                   processstr * pr,viewredrawstr * v)
{
 return(0);

 USE(picturep);
 USE(tr);
 USE(pr);
 USE(v);
}




static os_error * epsinfo(int object,int ftype,filestr * file,int * len)
{
 os_error          * err;
 picturestr        * picturep;
 fstat               fs;
 pxpreftagstaticstr  reftag;
 char                name[FILENAMELEN];

 err=NULL;

 *len=0;

 picturep=(picturestr*)fob_find(object,file);
 if(picturep->imagelen)
 {
  *len=picturep->imagelen;
 }
 else
 {
  pxreadtag((pxtagstr*)&reftag,PXPREF1TAG,object,file);
  if(reftag.len)
  {
   err=pxexpandpath(name,NITEM(name),reftag.name,file);
   if(!err)
   {
    err=fs_stat(name,&fs);
    if(!err && fs.object) *len=fs.length;
   }
  }
 }

 return(err);

 USE(ftype);
}




static os_error * epspxinfo(int object,int ftype,
                                       filestr * file,int * info)
{
 picturestr       * picturep;
 char               string[128];
 int                len;


 picturep=(picturestr*)fob_find(object,file);


 *info++=PT_LENGTH;
 *info++=3*sizeof(int);

 epsinfo(object,ftype,file,&len);
 *info++=len;

 strcpy(string,"EPS");

 len=strlen(string);
 len=(len+3)&(~0x3);
 *info++=PT_INFO;
 *info++=2*sizeof(int)+len;
 strcpy((char*)info,string);
 info+=(len/sizeof(int));

 *info++=PT_DIM;
 *info++=4*sizeof(int);
 *info++=picturep->abox.x1-picturep->abox.x0;
 *info++=picturep->abox.y1-picturep->abox.y0;

 *info++=PT_CLASS;
 *info++=8*sizeof(int);
 *info++=PT_VEC;

 *info++=PT_RGB;
 *info++=3;
 *info++=8;
 *info++=0;
 *info++=0;

 *info++=PT_END;
 *info++=2*sizeof(int);

 return(NULL);
}




static int epsshift(picturestr * picturep)
{
 int shift;
 hdrshift(picturep,&shift,NULL);
 return(shift);
}





static os_error * epsrefinit(char * name,int ftype,filestr * file,int picture)
{
 os_error    * err;
 epsparsestr   eps;
 int           ptype;
 int           transhandle;
 char          transname[FILENAMELEN];
 void        * eh;


 err=epsparseinit(&eps);
 if(!err)
 {
  ptype=fs_filetypefromextension(name);

  if(!err)
  {
   if(ptype!=ftype)
   {
    err=sc_create(&transhandle);
    if(!err)
    {
     err=xzstrncpyu(transname,sc_name(transhandle),NITEM(transname));

     cprintf("translate %x %x",ptype,ftype);

     if(!err) err=translatefile(name,-1,transname,ftype);
     if(!err)
     {
      err=initpicture(picture,file,0,transname,&eps);
     }
 
     eh=errsave(err);
     fs_delete(sc_name(transhandle));
     sc_remove(transhandle,err);
     err=errload(eh,NULL);
    }
   } 
   else
   {
             err=initpicture(picture,file,0,name,&eps);
    if(!err) err=epscheckincludedfiles(&eps,name);
   }
  }
  epsparsefinit(&eps);
 }

 return(err);
}









static os_error * epsrefload(char * name,char * expandedname,int ftype,int * object,filestr * file)
{
 os_error         * err;
 picturestr       * picturep;
 int                picture;
 pxpreftagstaticstr reftag;

// err=pxexpandpath(expandedname,name,file);

 cprintf("epsrefload");

 err=NULL;

 if(!err)
 {
  err=fob_create(sizeof(picturestr),file,OPICTURE,(fobjptr)&picturep);
  if(!err)
  {
   picture=picturep->key;
   picturep->ftype=ftype;
   fob_setbits(picture,file,OBJ_CCINSTANCE,(fobjptr)&picturep);

   err=epsrefinit(expandedname,ftype,file,picture);

   if(!err)
   {
    reftag.tag=PXPREF1TAG;
    xzstrncpyu(reftag.name,name,NITEM(reftag.name));
    reftag.len=sizeof(pxtagstr)+((strlen(reftag.name)+1+3)&(~0x3));

    err=pxwritetag((pxtagstr*)&reftag,picture,file);
    if(!err)
    {
     reftag.tag=PXPREF2TAG;
     xzstrncpyu(reftag.name,expandedname,NITEM(reftag.name));
     reftag.len=sizeof(pxtagstr)+((strlen(reftag.name)+1+3)&(~0x3));
     err=pxwritetag((pxtagstr*)&reftag,picture,file);
    }

    *object=picture;
   }

   if(err)
   {
    *object=0;
    fob_delete(picture,file);
   }
  }
 }

 return(err);
}




static os_error * epsembed(int ftype,int object,filestr * file)
{
 os_error           * err;
 picturestr         * picturep;
 pxpreftagstaticstr   reftag;
 char                 expname[FILENAMELEN];
 int                  hdrlen;
 epsobjectstr       * epsobj;
 epsparsestr          eps;
 char               * name;


 /* need name/hdr calculation */

 err=pxreadtag((pxtagstr*)&reftag,PXPREF1TAG,object,file);
 if(reftag.len)
 {
  err=pxexpandpath(expname,NITEM(expname),reftag.name,file);
  if(!err)
  {
   name=expname;
   
   if(!err)
   {
    hdrlen=sizeof(epsobjectstr)+strlen(name)+1;
    hdrlen=(hdrlen+3) & (~0x3);
    err=dxembedchunk(EPSSHIFT+hdrlen,object,file);
   }

//   cprintf("err0 %s",err?err->errmess:"");

   if(!err)
   {
    picturep=(picturestr*)fob_find(object,file);
    picturep->ftype=ftype;

    memset(picturep->data,0,EPSSHIFT);
    epsobj=(epsobjectstr*)(picturep->data+EPSSHIFT);
    memset(epsobj,0,sizeof(epsobjectstr));
    epsobj->objectlength=picturep->imagelen-(EPSSHIFT+hdrlen);
    epsobj->length=(hdrlen+epsobj->objectlength+3) & (~0x3);
    strcpy(epsobj->name,name);

//   cprintf("err1 %s",err?err->errmess:"");

    err=epsparseinit(&eps);
    if(!err)
    {
              err=initpicture(object,file,1,name,&eps);
     if(!err) err=epsembedincludedfiles(&eps,name,object,file);
     epsparsefinit(&eps);
    }

//   cprintf("err2 %s",err?err->errmess:"");

    if(err)
    {
     dxscrapimage(object,file);
    }
    else
    {
     pxtagsize(object,file,PXPREF1TAG,0);
     pxtagsize(object,file,PXPREF2TAG,0);
    }
   }

//   cprintf("err3 %s",err?err->errmess:"");

  }
 }

//   cprintf("err4 %s",err?err->errmess:"");

 return(err);
}



static os_error * epsrefreload(char * name,char * expname,int ftype,
                                                  int picture,filestr * file)
{
 os_error         * err;
 pxpreftagstaticstr reftag;


 cprintf("epsref *re*load");

 err=epsrefinit(expname,ftype,file,picture);
 if(!err)
 {
  dxscrapimage(picture,file);

  reftag.tag=PXPREF1TAG;
  strcpy(reftag.name,name);
  reftag.len=sizeof(pxtagstr)+((strlen(reftag.name)+1+3)&(~0x3));

  err=pxwritetag((pxtagstr*)&reftag,picture,file);
  if(!err)
  {
   reftag.tag=PXPREF2TAG;
   strcpy(reftag.name,expname);
   reftag.len=sizeof(pxtagstr)+((strlen(reftag.name)+1+3)&(~0x3));
   err=pxwritetag((pxtagstr*)&reftag,picture,file);
  }
 }

 return(err);
}







static os_error * epsproxy(int ftype,int object,filestr * file,
                                                 pxproxystr * proxy)
{

 USE(ftype);
 USE(object);
 USE(file);
 USE(proxy);

 return(NULL);

}


static os_error * epsremoveproxy(int ftype,int object,filestr * file)
{


 USE(ftype);
 USE(object);
 USE(file);

 return(NULL);

}









static impexstr epsstr=
{
 EPS,
 "EPS",

 IMPXLOAD|IMPXPICT,   /* flags       */
 epsimport,           /* import      */
 epsimportpost,       /* import post */
 epsexport,           /* export      */
 epsrender,           /* render      */
 epsdsexport,         /* dsexport    */
 epsdsimport,         /* dsimport    */

 pictintexport,       /* internal export */
 epsinfo,             /* exportinfo      */

 epscache,            /* cache */

 epsrendermem,        /* render memory */

 epsshift,            /* datashift */


 epspxinfo,           /* info */
 epsrefload,          /* reference load */
 epsproxy,            /* make proxy */
 epsremoveproxy,      /* remove proxy */

 epsembed,            /* embed referenced image */
 epsrefreload,        /* reload                 */
 NULL,                /* x2                     */
 NULL,                /* x3                     */


};



static impexstr psstr=
{
 PSCRIPT,
 "PostScript",

 IMPXLOAD|IMPXPICT,   /* flags       */
 epsimport,           /* import      */
 epsimportpost,       /* import post */
 epsexport,           /* export      */
 epsrender,           /* render      */
 epsdsexport,         /* dsexport    */
 epsdsimport,         /* dsimport    */

 pictintexport,       /* internal export */
 epsinfo,             /* exportinfo      */

 epscache,            /* cache */

 epsrendermem,        /* render memory */

 epsshift,            /* datashift */


 epspxinfo,           /* info */
 epsrefload,          /* reference load */
 epsproxy,            /* make proxy */
 epsremoveproxy,      /* remove proxy */

 epsembed,            /* embed referenced image */
 epsrefreload,        /* reload                 */
 NULL,                /* x2                     */
 NULL,                /* x3                     */


};









static os_error * drawevent(int eventn,int userhandle,void * data,int data1)
{
 impexregister(&epsstr);
 impexregister(&psstr);
 remevent(EVENT_DRAWINIT,drawevent,0);

 return(NULL);

 USE(eventn);
 USE(userhandle);
 USE(data);
 USE(data1);
}








extern "C" __declspec(dllexport) os_error * entrypoint(char * path)
{ 
 os_error * err;

          err=addevent(EVENT_DRAWINIT,drawevent,0);
 if(!err) err=adderrortable(errortable,&erroro1);
 if(!err) err=addevent(EVENT_PRCACHE0,(eventfn)epsprcache0,0);


 return(err);

 USE(path);
}





extern "C" BOOL __stdcall _DllMainCRTStartup( 
    HINSTANCE hinstDLL,     // handle to DLL module 
    DWORD fdwReason,        // reason for calling function 
    LPVOID lpvReserved)     // reserved 
{
 
 modulehandle=(HMODULE)hinstDLL;
 return(TRUE);
}
