/*->c.tw */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "transform.h"
#include "xprocess.h"


#include "err.h"
#include "fsx.h"
#include "task.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "etc.h"
#include "bf.h"
#include "colour.h"
#include "deb.h"
#include "dbhi.h"
#include "bitrev.h"
#include "ram.h"
#include "scrap.h"
#include "filetype.h"

#include "constants.h"
#include "str.h"

#include "obj.h"
#include "object.h"
#include "chain.h"
#include "view.h"
#include "viewr.h"
#include "file.h"
#include "select.h"

#include "txstr.h"
#include "mx.h"
#include "pxstr.h"
#include "pxbmp.h"
#include "pict.h"
#include "draw.h"


#include "twain.h"
#include "dca_glue.h"        // for function prototypes of glue code
#include "dca_type.h"        // contains function protos for this module
#include "dca_acq.h"         // contains buffered glue code function prototypes
#include "dca_app.h"



#include "tw.h"



/****************************************************************************/



/*****************************************************************************/

static filestr         * acqfile;
static viewstr         * acqview;




/* open select source dialogue */

os_error * twainselect(void)
{
 os_error * err;
 viewstr  * view;

 err=NULL;

 view=getmenuview();
 if(view)
 {
  err=TWSelectSource(view->handle);
 }

 return(err);
}

/****************************************************************************/
/* code for acquiring images */

static TWAcquireblockstr acqblock;

static int acqpict;


#define DIBSHIFT 14


static os_error * twload(int object,int type)
{
 os_error   * err;
 posnstr      posn;
 pictboxstr * pictp;
 int          page;
 int          section;


 bmpinitpicture(object,type,acqfile);

 pictp=(pictboxstr*)fob_find(acqpict,acqfile);
 pagesectionof((chainobject*)pictp,acqfile,&page,&section);

 err=modfile(acqfile,section);
 if(!err)
 {
 	memset(&posn,0,sizeof(posnstr));
 	posn.branch=acqview->branch;
 	posn.section=section;
 	posn.page=page;
 	posn.file=acqfile;
 	posn.view=acqview;
 	posn.object=acqpict;

  err=pictloadsub(BMP,object,&posn);
 }

 return(err);
}





/****************************************************************************/

static int twsh;


static os_error * acqfilesetup(int code)
{
 os_error * err;

 err=NULL;

 if(code==1)
 {
  if(!twsh)  err=sc_create(&twsh);
  acqblock.filename=sc_namex(twsh);
 }
 else
 if(code==0)
 {
  err=sc_remove(twsh,NULL);
  twsh=0;
 }

 return(err);
}



static os_error * acqfilexfer(int last)
{
 os_error * err;
 int        object;

 err=NULL;

 if(last==TWLASTBLOCK)
 {
           err=dxloadchunk(acqblock.filename,0,&object,acqfile);
  if(!err) err=twload(object,BMP);
 }

 return(err);
}



/****************************************************************************/


// native transfer - pointer to dib

static os_error * acqsprite(int last,int length)
{
 os_error   * err;
 picturestr * picturep;
 int          picturek;

 err=NULL;

 if(last==TWLASTBLOCK)
 {
  err=fob_create(length+sizeof(picturestr)+DIBSHIFT,acqfile,OPICTURE,(fobjptr)&picturep);
  if(!err)
  {
   picturek=picturep->key;
   memcpy(picturep->data+DIBSHIFT,acqblock.memory.native.lpdib,length);
   picturep->imagelen=length+DIBSHIFT;

   fob_setbits(picturek,acqfile,OBJ_CCINSTANCE,(fobjptr)&picturep);

   err=twload(picturek,DIB);
  }
 }


 return(err);
}


/****************************************************************************/
/* code for direct memxfers */

static int       memline;
static int       memplanar;
static int       memplane;

static int       fxpix;
static int       fypix;
static int       fbpp;
static int       fxdpi;
static int       fydpi;
static int       fwwidth;
static int       fshift;
static int       fobject;



typedef struct xbmphdrstr
{
 int        cbsize;
 short int  xhotspot;
 short int  yhotspot;
 int        offbits;    
} xbmphdrstr;




#define QUADFROMPAL(a) ( (((a) & 0xFF00)<<8) | (((a) & 0xFF0000)>>8) | (((a)>>24) & 0xFF) )

static void acqmemxferpalette(int * paltab,int nc)
{
 picturestr       * picturep;
 int                i;
 char             * data;
 int                pal[256]; 

 for(i=0;i<nc;i++) pal[i]=QUADFROMPAL(paltab[i]);

 picturep=(picturestr*)fob_find(fobject,acqfile);
 data=(picturep->data+2+sizeof(xbmphdrstr)+sizeof(BITMAPINFOHEADER));

 memcpy(data,pal,nc*sizeof(int));

// cprintf("xfer pal %d pal[2]=%x pal[4]=%x",nc,pal[2],pal[4]);

}


static os_error * acqsetmemxfer(int xpix,int ypix,int bpp,int xres,
                                                          int yres,int planar)
{
 os_error         * err;
 int                length;
 BITMAPINFOHEADER   bmi;
 picturestr       * picturep;
 xbmphdrstr         hdr;


	cprintf("bpp %d",bpp);

 err=NULL;

 memline=0;
 memplanar=planar;
 memplane=0;


 if(!err)
 {
  fxpix=xpix;
  fypix=ypix;

  fbpp=bpp;
  fxdpi=xres;
  fydpi=yres;

  fwwidth=((bpp*fxpix+31) & ~0x1F)>>3;

  length=2+sizeof(xbmphdrstr)+sizeof(BITMAPINFOHEADER);
  if(bpp<=8) length+=4*(1<<bpp);
  fshift=length;
  length+=fypix*fwwidth;

  hdr.cbsize=length;
  hdr.xhotspot=0;
  hdr.yhotspot=0;
  hdr.offbits=fshift;

  bmi.biSize=sizeof(BITMAPINFOHEADER);
  bmi.biWidth=fxpix;
  bmi.biHeight=fypix;
  bmi.biXPelsPerMeter=(fxdpi*10000)/254;
  bmi.biYPelsPerMeter=(fydpi*10000)/254;
  bmi.biBitCount=(unsigned short)bpp;
  bmi.biPlanes=1;
  bmi.biCompression=BI_RGB;
  bmi.biSizeImage=0;
  bmi.biClrImportant=0;
  bmi.biClrUsed=0;

//  cprintf("bisize %d",sizeof(BITMAPINFOHEADER));



  fob_create(length+sizeof(picturestr),acqfile,OPICTURE,(fobjptr)&picturep);
  fobject=picturep->key;
  picturep->imagelen=length;
  memcpy(picturep->data+2+sizeof(xbmphdrstr),&bmi,bmi.biSize);
  memcpy(picturep->data+2,&hdr,sizeof(xbmphdrstr));
  picturep->data[0]='B';
  picturep->data[1]='M';
  fob_setbits(fobject,acqfile,OBJ_CCINSTANCE,(fobjptr)&picturep);


  if(!err)
  {
//   imsetgreypal(acqimage,bpp);
  }
 }

 return(err);
}


/* just received a chunk */

static os_error * acqmemxfer(int size,int last,int bpr,int rows)
{
 os_error   * err;
 int        * idata;
 char       * p;
 int          i;
 char       * q;
 int          bytes;
 picturestr * picturep;



 err=NULL;

 for(i=0;i<rows;i++)
 {
/*  if(memline>=acqimage->ypix) dprintf(0,"too many lines");
  if(bpr>acqimage->wwidth*4) dprintf(0,"too many bytes %d %d",bpr,acqimage->wwidth*4); */

  if(memline>=fypix && memplanar) 
  {
   memplane++;
   memline=0;
  }

  if(memline>=fypix) break;
  if(memplane>=3)    break;

  picturep=(picturestr*)fob_find(fobject,acqfile);
  idata=(int*)(picturep->data+fshift+(fypix-1-memline)*fwwidth);

  p=((char*)acqblock.memory.memxfer.buffer)+bpr*i;

  if(memplanar)
  {
   bytes=MIN(bpr,fxpix);
   q=(char*)idata;
   q+=memplane;
   while(bytes--)
   {
    *q=*p++;
    q+=3;
   }
  }
  else
  {
   memcpy(idata,p,MIN(bpr,fwwidth));
			if(fbpp==24)
			{
				p=(char*)idata;
				q=((char*)idata)+2;
    bytes=MIN(bpr,fxpix);
				while(bytes--)
				{
					SWOP((*p),(*q));
					p+=3;
					q+=3;
				}
			}
  }

  memline++;
 }


 if(!err)
 {
  if(last==1)
  {
   err=twload(fobject,BMP);
  }
  else
  if(last==-1) /* failed */
  {
   fob_delete(fobject,acqfile);
   fobject=0;
  }
 }


 return(err);

 USE(size);
}



/*****************************************************************************/




static os_error * acqend(void)
{
 if(acqblock.transfertype==TWSX_MEMORY)
 {
  if(acqblock.memory.memxfer.buffer)
                        sfree((flex_ptr)&acqblock.memory.memxfer.buffer);

 }

 acqview=NULL;
 acqfile=NULL;

 return(NULL);
}



static os_error * acqinfo(int xpix,int ypix,int bpp,int xres,
                                                    int yres,int planar)
{
 if(acqblock.transfertype==TWSX_MEMORY)
 {
  acqsetmemxfer(xpix,ypix,bpp,xres,yres,planar);
 }

 return(NULL);
}

static os_error * acqpalette(int * paltab,int nc)
{

// cprintf("acq palette %d",nc);

 if(acqblock.transfertype==TWSX_MEMORY)
 {
  acqmemxferpalette(paltab,nc);
 }

 return(NULL);
}



static os_error * acqtransfer(int size,int last,int bpr,int rows)
{
 os_error * err;

 err=NULL;

 if(acqblock.transfertype==TWSX_NATIVE) err=acqsprite(last,size);
 else
 if(acqblock.transfertype==TWSX_MEMORY) err=acqmemxfer(size,last,bpr,rows);
 else
 if(acqblock.transfertype==TWSX_FILE)   err=acqfilexfer(last);

 return(err);
}




static os_error * acqalloc(int size)
{
 os_error * err;

 err=NULL;

 if(acqblock.transfertype==TWSX_MEMORY)
 {
  err=salloc((flex_ptr)&acqblock.memory.memxfer.buffer,size);

 }

 return(err);
}




static os_error * closeevent(int eventn,int userhandle,void * data,int data1)
{
 viewstr * view;
 filestr * file;

 if(eventn==EVENT_FCLOSE)
 {
  file=(filestr*)data;

  if(file==acqfile && file)
  {
   TWTerminate();
   acqfile=NULL;
   acqview=NULL;
  }
 }
 else
 if(eventn==EVENT_VCLOSE)
 {
  view=(viewstr*)data;

  if(view==acqview && view)
  {
   TWTerminate();
   acqfile=NULL;
   acqview=NULL;
  }
 }

 return(NULL);

 USE(userhandle);
 USE(data1);
}



os_error * twainacquire(void)
{
 os_error   * err;
 pictboxstr * pictp;
 int          page;
 int          section;

 err=NULL;

 acqview=getmenuview();
 acqfile=acqview->file;


 if(nselected(&acqpict,acqfile))
 {
  pictp=(pictboxstr*)fob_find(acqpict,acqfile);
  if(pictp->type!=OPICTBOX)
  {
   pagesectionof((chainobject*)pictp,acqfile,&page,&section);
   err=txconvertable(acqpict,section,acqfile);
  }

  if(!err)
  {
   TWAcquireDefault(&acqblock);


   if(selection)  /* there is a selected area that we want */
   {
//  acqblock.setarea=1;
//  acqblock.area.x0=sbox.x;
//  acqblock.area.y1=sbox.y;
//  acqblock.area.x1=sbox.x+sbox.w;
//  acqblock.area.y0=sbox.y+sbox.h;
   }


   acqblock.transfertype=TWSX_MEMORY;


   acqblock.settransfertype=1;

   acqblock.TWend=acqend;
   acqblock.TWalloc=acqalloc;
   acqblock.TWtrans=acqtransfer;
   acqblock.TWinfo=acqinfo;
   acqblock.TWfilesetup=acqfilesetup;

   acqblock.TWpal=acqpalette;
 
   err=TWAcquire(acqview->handle,1,1,&acqblock);

  }
 }

 return(err);
}


/*****************************************************************************/

static os_error * twainterminate(int eventn,int userhandle,
                                                        void * data,int data1)
{

 TWTerminate();

 return(NULL);

 USE(eventn);
 USE(userhandle);
 USE(data);
 USE(data1);
}


os_error * twaininit(void)
{
 TW_IDENTITY AppIdentity;

 AppIdentity.Id=0;/* init to 0, but Source Manager will assign real value */

 AppIdentity.Version.MajorNum=1;
 AppIdentity.Version.MinorNum=0;
 AppIdentity.Version.Language=TWLG_ENG;
 AppIdentity.Version.Country=TWCY_UNITEDKINGDOM;
 strcpy(AppIdentity.Version.Info,wimpt_programname());

 AppIdentity.ProtocolMajor=TWON_PROTOCOLMAJOR;
 AppIdentity.ProtocolMinor=TWON_PROTOCOLMINOR;
 AppIdentity.SupportedGroups=DG_IMAGE | DG_CONTROL;

 strcpy(AppIdentity.Manufacturer,"David Pilling");
 strcpy(AppIdentity.ProductFamily,wimpt_programname());
 strcpy(AppIdentity.ProductName,wimpt_programname());

 /* pass app particulars to glue code */
 
// TWInitialize(&AppIdentity,(HWND)taskhandle);
 TWInitialize(&AppIdentity,0);

 addevent(EVENT_ENDPROGRAM,twainterminate,0);

 addevent(EVENT_FCLOSE,closeevent,0);
 addevent(EVENT_VCLOSE,closeevent,0);

 return(NULL);
}

